import hashlib
import types
from typing import Any, Callable, Tuple

import pandas as pd
import pyarrow as pa

import federleicht.args as args
from federleicht.config import CACHE

__salt__: bytes = None


def salt() -> bytes:
    """Generate a salt based on the main dependency versions.

    Returns:
        bytes: The first 8 characters of the package version encoded as bytes.
    """

    global __salt__

    if __salt__ is None:

        dependencies = {
            "pandas": pd.__version__,
            "pyarrow": pa.__version__,
        }

        versions = ";".join(f"{key}={value}" for key, value in dependencies.items())

        __salt__ = hashlib.blake2s(
            versions.encode(),
            digest_size=8,
            salt=CACHE.version.encode()[:8],
        ).digest()

    return __salt__


def hash_wrapped(
    function: types.FunctionType,
    arguments: Tuple[Any, ...],
) -> hashlib.blake2s:
    """Generate a BLAKE2s hash for all arguments and also the functions byte-code.

    Args:
        function (types.FunctionType): wrapped function.
        arguments (Tuple[Any, ...]): *args and **kwargs of wrapped function

    Returns:
        hashlib.blake2s: The BLAKE2s hash object.
    """

    binarydata = args.unique(*arguments).digest()

    hash = hashlib.blake2s(
        binarydata,
        digest_size=CACHE.digest,
        salt=salt(),
    )

    hash.update(function.__qualname__.encode())
    hash.update(function.__code__.co_code)

    return hash


def function(
    function: types.FunctionType,
    arguments: Tuple[Any, ...],
    pepper: Callable[[], bytes] = None,
) -> str:
    """Generate a BLAKE2s hash for all arguments and also the functions byte-code.

    Args:
        function (types.FunctionType): wrapped function.
        arguments (Tuple[Any, ...]): *args and **kwargs of wrapped function
        pepper (Callable[[], bytes], optional): A function that returns a salt.
            Defaults to None.

    Returns:
        str: The hexadecimal digest of the BLAKE2s hash.
    """

    hash = hash_wrapped(function, arguments)

    if callable(pepper):
        hash.update(pepper())

    return hash.hexdigest()



"""
.. include:: ../README.md
"""

from federleicht.cache import clear_cache, delete_cache, from_cache
from federleicht.config import __version__  # noqa: F401
from federleicht.dataframe import cache_dataframe

__all__ = [
    "from_cache",
    "clear_cache",
    "delete_cache",
    "cache_dataframe",
]



import importlib.metadata
from collections import namedtuple

__version__ = importlib.metadata.version(__package__)

CacheConfig = namedtuple(
    "CacheConfig",
    [
        "version",
        "dir",
        "digest",
        "expires",
        "attrs",
    ],
)

CACHE = CacheConfig(
    version=__version__,
    dir=".pandas_cache",
    digest=16,
    expires="seconds",
    attrs=".json",
)
"""
CACHE configuration.

Attributes:
    version (str): The version of the package.
    dir (str): The directory where the cache is stored.
    digest (int): The digest size for hashing.
    expires (str): The expiration time unit for the cache.
    attrs (str): The file extension for attribute storage.
"""



"""
The `unique` function generates a determanistic hash for all positional and keyword
arguments.

The hash is calculated with over the json string representation of all arguments.
The order of the keyword arguments is not relevant, all arguments must be serializable.

For `os.PathLike` objects a tuple with the resolved path, size and modification time is
used for hashing to detect if the file changed.

For `numpy.ndarray`, `pandas.DataFrame` and `pandas.Series` a tuple of the type and the
binary representation is used for hashing to detect if the data changed.
"""

import io
import json
import os
import pathlib
import types
from abc import ABC, abstractmethod
from datetime import datetime
from decimal import Decimal
from fractions import Fraction
from types import MappingProxyType
from typing import Any, Dict

import pandas as pd

try:
    from xxhash import xxh128 as hash  # type: ignore
except ModuleNotFoundError:
    from hashlib import md5 as hash


class _Hash(ABC):  # pragma: no cover
    """
    Abstract base class for hashing algorithms to use as type hint for the `hash`
    function.
    """

    @abstractmethod
    def __init__(self, readable: bytes) -> None:
        pass

    @abstractmethod
    def update(self, data: bytes) -> None:
        """
        Update the hash object with the bytes in data. Repeated calls
        are equivalent to a single call with the concatenation of all
        the arguments.
        """
        pass

    @abstractmethod
    def digest(self) -> bytes:
        """
        Return the digest of the bytes passed to the update() method
        so far as a bytes object.
        """
        pass

    @abstractmethod
    def hexdigest(self) -> str:
        """
        Like digest() except the digest is returned as a string
        of double length, containing only hexadecimal digits.
        """
        pass

    @abstractmethod
    def copy(self) -> "_Hash":
        """
        Return a copy (clone) of the hash object. This can be used to
        efficiently compute the digests of datas that share a common
        initial substring.
        """
        pass


def representation(obj: Any) -> str:
    """
    Return an unique string representation of the object.

    Raises a TypeError if the given object is not an immutable type.
    """
    immutable_types = (
        datetime,
        int,
        float,
        complex,
        str,
        tuple,
        frozenset,
        bytes,
        Decimal,
        Fraction,
        MappingProxyType,
    )

    if not isinstance(obj, immutable_types):
        raise TypeError(
            f"Object of type {type(obj).__qualname__} is not immutable: {obj}"
        )

    return str(obj)


def json_encoder(obj: Any) -> str:
    """
    Encode an object to a unique string representation, for mutable types it can result
    in returning a hexdigest of the object.
    """

    # pathlib.Path
    if isinstance(obj, os.PathLike):
        p = pathlib.Path(obj)
        return str(
            (
                p.resolve().as_posix(),
                p.stat().st_size,
                p.stat().st_mtime,
            )
        )

    # pandas.DataFrame
    if hasattr(obj, "to_feather"):
        buffer = io.BytesIO()
        obj.to_feather(buffer)
        return str(
            (
                type(obj),
                hash(buffer.getvalue()).hexdigest(),
                unique(**obj.attrs).hexdigest(),
            )
        )

    if isinstance(obj, pd.Series):
        return json_encoder(obj.to_frame())

    # numpy.ndarray
    if hasattr(obj, "tobytes"):
        return str(
            (
                type(obj),
                hash(obj.tobytes()).hexdigest(),
            )
        )

    if isinstance(obj, types.FunctionType):
        return str(
            (
                obj.__qualname__,
                hash(obj.__code__.co_code).hexdigest(),
            )
        )

    return representation(obj)


def dumps(
    *args: Any,
    **kwargs: Dict[str, Any],
) -> str:
    """
    Generate a JSON string for all arguments and keyword arguments.
    """

    data = json.dumps(
        (args, kwargs),
        sort_keys=True,
        default=json_encoder,
        indent=4,
    )

    return data


def unique(
    *args: Any,
    **kwargs: Dict[str, Any],
) -> _Hash:
    """
    Generate unique hash for all positional and keyword arguments. The order of the
    keyword arguments is not relevant.
    """

    data = dumps(*args, **kwargs)

    return hash(data.encode())


__all__ = [
    "unique",
    "hash",
]



import json
from typing import Any, Callable, Dict

import pandas as pd
from pathlibutil import Path

import federleicht.args as args
from federleicht.config import CACHE


def lock(mydict: Dict, **kwargs) -> str:
    """
    Create a hash for the dict, skip all keys which are not of type int, float, str,
    bool, or None.

    All value types must be serializable to JSON.
    """

    skip_keys = (int, float, str, bool, type(None))

    cleandict = {k: v for k, v in mydict.items() if type(k) in skip_keys}

    attrs = json.dumps(cleandict, sort_keys=True, **kwargs)

    return args.hash(attrs.encode()).hexdigest()


def save(
    df: pd.DataFrame,
    filename: Path,
    default: Callable[[Any], str] = None,
) -> Path:
    """
    Dump the DataFrame attributes and their hash into a JSON file.

    Creates a JSON file only when `df.attrs` is not empty.

    Args:
        df (pd.DataFrame): The DataFrame whose attributes are to be dumped.
        file (str): The file path where the JSON file will be created.
        default (Callable[[Any], str]): A function to convert non-serializable objects.

    Returns:
        Path: The path to the created JSON file, or None if df.attrs is empty.
    """

    if not df.attrs:
        return None

    file: Path = filename.with_suffix(CACHE.attrs)

    file.write_text(
        json.dumps(
            {
                "attrs": df.attrs,
                "lock": lock(df.attrs, default=default),
            },
            default=default,
            indent=4,
        ),
    )

    return file


def restore(df: pd.DataFrame, filename: Path, **kwargs) -> pd.DataFrame:
    """
    Verify and restore the DataFrame attributes from a JSON file.

    Args:
        df (pd.DataFrame): The DataFrame to which the attributes will be loaded.
        file (str): The file path from where the JSON file will be read.
        **kwargs: Additional arguments for `json.loads`.

    Returns:
        pd.DataFrame: The DataFrame with updated attributes.
    """

    file: Path = filename.with_suffix(CACHE.attrs)

    try:
        cached = json.loads(file.read_text(), **kwargs)
    except FileNotFoundError:
        return df

    try:
        attrs = cached["attrs"]

        if lock(attrs) == cached["lock"]:
            df.attrs.update(attrs)
    except KeyError:
        pass

    return df


__all__ = [
    "save",
    "restore",
]



import re

import pandas as pd
from pathlibutil import Path

from federleicht.config import CACHE


def from_cache(df: pd.DataFrame) -> bool:
    """
    Check if a DataFrame was loaded from the cache.

    Args:
        df (pd.DataFrame): The DataFrame to check.

    Returns:
        bool: True if the DataFrame was loaded from the cache, False otherwise.
    """
    return df.attrs.get("from_cache", None) is not None


def delete_cache(cache_file: str) -> None:
    """
    Delete a cache file and its optional attributes file.

    Args:
        filename (str): The name of the cache file to delete.

    Example:
        >>> delete_cache("cachefile")
    """

    cache = Path(cache_file)

    cache.delete(missing_ok=True)
    cache.with_suffix(CACHE.attrs).delete(missing_ok=True)


def clear_cache(cache_dir: str = CACHE.dir, **kwargs) -> int:
    """
    Clear all cache files from the cache directory.

    Args:
        cache_dir (str): The directory where the cache files are stored.
        **kwargs: argument for `datetime.timedelta` to determine expiration time.

    Returns:
        int: The number of errors encountered while clearing the cache.

    Example:
        >>> clear_cache(weeks=2)
        0
    """

    if not kwargs:
        is_expired = lambda file: True  # noqa
    else:
        is_expired = lambda file: file.is_expired(**kwargs)  # noqa

    regex = re.compile(f"[a-f0-9]{{{CACHE.digest}}}", re.IGNORECASE)

    error = 0
    files = (file for file in Path(cache_dir).iterdir() if file.is_file())

    for file in files:
        if not regex.fullmatch(file.name):
            continue

        if not is_expired(file):
            continue

        try:
            delete_cache(file)
        except Exception:
            error += 1

    return error



from functools import wraps
from typing import Any, Callable, Dict, Union

import pandas as pd
from pathlibutil import Path

import federleicht.attrs as attrs
import federleicht.hash as hash
from federleicht.cache import delete_cache
from federleicht.config import CACHE


def is_expired(file: Path, expires: Union[int, Dict[str, Any]]) -> bool:
    """Check if a file is expired.

    Args:
        file (Path): The file to check.
        expires (Union[int, Dict[str, Any]]): The expiration time in seconds or a
            dictionary containing kwargs for `datetime.timedelta`.

    Example:
        ```python
        file_to_cache = pathlibutil.Path(__file__)
        hours_to_cache: int = 90
        if is_expired(file_to_cache, expires={"hours": hours_to_cache}):
            file_to_cache.unlink()
        ```

    Returns:
        bool: True if the file is expired or expires is None, False otherwise.
    """

    if expires is None:
        return False

    if isinstance(expires, int):
        timedelta = {CACHE.expires: expires}
    elif isinstance(expires, dict):
        timedelta = expires
    else:
        raise TypeError(f"Invalid type expires: {type(expires)}. Must be int or dict.")

    return file.is_expired(**timedelta)


def cache_dataframe(
    func: Callable = None,
    *,
    cache_dir: str = CACHE.dir,
    expires: Union[int, Dict[str, int]] = None,
    cache_attrs: bool = False,
    pepper: Callable[[], bytes] = None,
):
    """
    Decorator to cache the result of a function that returns a pandas DataFrame.

    To reliable cache a DataFrame the decorated function must always return the same
    DataFrame for the same arguments!
    All arguments of the decorated function must be pickleable!

    The cache expires when:
    - the expiration time is reached
    - the decorated function changes
    - the arguments to the decorated function change
    - the pandas_cache version changes

    If the cache exists and is not expired, it loads the DataFrame from the cache.
    Otherwise, it calls the function, caches the result, and then returns the DataFrame.

    Args:
        func (callable, optional): The function to be decorated. Defaults to None.
        cache_dir (str, optional): Directory where the cache files will be stored.
            Defaults to CACHE.dir.
        expires (Union[int, Dict[str, int]], optional): Expiration time for the cache.
            Can be an integer representing the number of seconds or a dictionary with
            time units (e.g., {"hours": 1}). Defaults to None.
        cache_attrs (bool, optional): Whether to cache DataFrame attributes. These must
            all be serializable. Defaults to False.
        pepper (callable, optional): A function that returns a salt to spice up the
            hash of the cache. Defaults to None.

    Returns:
        callable: The wrapped function with caching functionality.

    Raises:
        TypeError: If the `expires` argument is not an int or dict.

    Example:
        ```python
        @cache_dataframe(expires={"hours": 1}, cache_attrs=True)
        def create_dataframe(*args, **kwargs):
            return pd.DataFrame(*args, **kwargs)
        ```
    """

    if func is None:
        return lambda f: cache_dataframe(
            f,
            expires=expires,
            cache_dir=cache_dir,
            cache_attrs=cache_attrs,
            pepper=pepper,
        )

    @wraps(func)
    def wrapper(*args, **kwargs):

        lock: str = hash.function(func, (args, kwargs), pepper)
        cache: Path = Path(cache_dir).joinpath(lock)

        try:
            if is_expired(cache, expires):
                delete_cache(cache)
                raise FileNotFoundError

            df = pd.read_feather(cache)
            df.attrs["from_cache"] = cache

            if cache_attrs is True:
                df = attrs.restore(df, cache)

        except FileNotFoundError:
            df: pd.DataFrame = func(*args, **kwargs)

            cache.parent.mkdir(parents=True, exist_ok=True)
            df.to_feather(cache)

            if cache_attrs is True:
                attrs.save(df, cache)

        return df

    return wrapper
